import dataclasses
import enum
from functools import cached_property
from typing import Optional, Any, List, Callable, Dict


@dataclasses.dataclass
class InstructionFeature:
    instruction_id: int
    name: str
    value: Any


@dataclasses.dataclass
class Question:
    question: str
    answer: str


@dataclasses.dataclass
class Instruction:
    id: int
    instruction: str
    output: str
    input: Optional[str]
    features: Optional[List[InstructionFeature]] = dataclasses.field(
        default=None,
    )

    @cached_property
    def text(self):
        return (
            self.instruction
            + "\n"
            + (self.input if self.input else "")
            + " \n"
            + self.output
        )

    @property
    def qa(self) -> Optional[Question]:
        question = self.instruction + ((" \n" + self.input) if self.input else "")
        answer = self.output

        return Question(question=question, answer=answer)

    def is_question(self) -> bool:
        return self.instruction.strip().endswith("?")

    def to_dict(self, include_features: bool = False):
        return {
            "id": self.id,
            "instruction": self.instruction,
            "input": self.input,
            "output": self.output,
        } | (
            {
                f"feature_{feature.name}": feature.value
                for feature in self.features or dict()
            }
            if include_features
            else {}
        )

    def to_conversation_dict(self) -> List[Dict]:
        return [
            {"role": "user", "content": f"{self.instruction} \n {self.input}"},
            {"role": "assistant", "content": self.output},
        ]

    @cached_property
    def _features_dict(self):
        return {f.name: f for f in self.features}

    def get_feature(self, name):
        return self._features_dict[name]


@dataclasses.dataclass
class CustomExtractorDefinition:
    name: str
    f: Callable[[Instruction], InstructionFeature]


class SubsetChoiceMethod(enum.Enum):
    RANDOM = "RANDOM"
    VARIABILITY_FACTOR = "VARIABILITY_FACTOR"


class EmbeddingComputationStrategy(enum.Enum):
    TRUNCATE = "TRUNCATE"


class DatasetError(BaseException):
    pass



import sys

import logging
import logging.config

from typing import Optional

import coloredlogs

from elemelek.settings import LOGGING_FORMAT, LOGGING_LEVEL, LOGGING_PLAIN


def add_handler(logger, handler, level, fmt):
    handler.setLevel(level)
    formatter = logging.Formatter(fmt)
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def get_logger(
    name: str = "",
    level: str = LOGGING_LEVEL,
    file_handler_path: str = None,
    fmt: str = LOGGING_FORMAT,
) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.setLevel(level)
    if logger.hasHandlers():
        return logger

    if file_handler_path:
        add_handler(logger, logging.FileHandler(file_handler_path), level, fmt)
    else:
        if LOGGING_PLAIN:
            add_handler(logger, logging.StreamHandler(sys.stdout), level, fmt)
        else:
            coloredlogs.install(level=level, logger=logger, fmt=fmt)

    logger.propagate = False

    return logger


class SelfLogging:
    _logger: Optional[logging.Logger] = None

    def _create_logger(self):
        return get_logger(self.__class__.__name__)

    @property
    def logger(self):
        if not self._logger:
            self._logger = self._create_logger()
        return self._logger

    @logger.setter
    def logger(self, value):
        self._logger = value

    def debug(self, msg, **kwargs):
        self.logger.debug(msg, **kwargs)

    def info(self, msg, **kwargs):
        self.logger.info(msg, **kwargs)

    def error(self, msg, **kwargs):
        self.logger.error(msg, **kwargs)

    def warn(self, msg, **kwargs):
        self.logger.warning(msg, **kwargs)

    def exception(self, msg, **kwargs):
        self.logger.exception(msg, **kwargs)



import random

import numpy as np

from elemelek.logging import SelfLogging


class OptimalSubsetGeneticAlgorithm(SelfLogging):
    def __init__(
        self,
        distance_matrix: np.ndarray,
        target_distance: float,
        population_size: int,
        sample_size: int,
        generations: int,
        mutation_rate: float,
    ):
        self.distance_matrix = distance_matrix
        self.target_distance = target_distance
        self.population_size = population_size
        self.sample_size = sample_size
        self.generations = generations
        self.mutation_rate = mutation_rate

    def _init_solution(self) -> np.ndarray:
        solution = np.zeros(len(self.distance_matrix), dtype=np.uint8)
        indices = np.random.choice(
            np.arange(len(solution)), self.sample_size, replace=False
        )
        solution[indices] = 1

        assert len(solution.nonzero()[0]) == self.sample_size
        return solution

    def _eval(self, solution: np.ndarray) -> float:
        submatrix = self.distance_matrix[
            np.ix_(solution.nonzero()[0], solution.nonzero()[0])
        ]

        return abs(np.median(submatrix) - self.target_distance)

    def _cross(self, solution_a: np.ndarray, solution_b: np.ndarray) -> np.ndarray:
        non_zero_indices = np.nonzero(solution_a + solution_b)[0]
        new_solution = np.zeros_like(solution_a, dtype=np.uint8)
        chosen = np.random.choice(
            non_zero_indices, size=self.sample_size, replace=False
        )

        new_solution[chosen] = 1
        assert len(new_solution.nonzero()[0]) == self.sample_size
        return new_solution

    def _mutate(self, solution: np.ndarray) -> np.ndarray:
        mutation_indices = np.random.choice(
            np.arange(len(solution)), size=self.sample_size, replace=False
        )
        for index in mutation_indices:
            solution[index] = 1 - solution[index]

        ones = np.where(solution == 1)[0]
        zeros = np.where(solution == 0)[0]
        if len(ones) > self.sample_size:
            drop_indices = np.random.choice(
                ones, size=len(ones) - self.sample_size, replace=False
            )
            solution[drop_indices] = 0
        elif len(ones) < self.sample_size:
            add_indices = np.random.choice(
                zeros, size=self.sample_size - len(ones), replace=False
            )
            solution[add_indices] = 1

        assert len(solution.nonzero()[0]) == self.sample_size
        return solution

    def optimize(self):
        population = [self._init_solution() for _ in range(self.population_size)]
        best_solution = None
        best_fitness = np.inf

        for generation in range(self.generations):
            fitness = np.array([self._eval(sol) for sol in population])
            min_index = np.argmin(fitness)
            if fitness[min_index] < best_fitness:
                best_fitness = fitness[min_index]
                best_solution = population[min_index]

            new_population = []
            while len(new_population) < self.population_size:
                parents = random.sample(population, 2)
                offspring = self._cross(parents[0], parents[1])
                if random.random() < self.mutation_rate:
                    offspring = self._mutate(offspring)
                new_population.append(offspring)
            population = new_population
        assert len(best_solution.nonzero()[0]) == self.sample_size
        return best_solution.nonzero()[0]



import hashlib
import math

from collections import Counter
from typing import List, Dict, Sequence

import language_tool_python

from elemelek.model import InstructionFeature
from elemelek.settings import LANGUAGE_TOOL_CHECK


def calculate_file_md5(file_path: str, block_size=8192):
    hasher = hashlib.md5()
    with open(file_path, "rb") as f:
        for block in iter(lambda: f.read(block_size), b""):
            hasher.update(block)
    return hasher.hexdigest()


def calculate_text_chunk_md5(texts: List[str]):
    hasher = hashlib.md5()
    for text in texts:
        hasher.update(text.encode())
    return hasher.hexdigest()


def divide_dict(d: dict, n: int) -> List[List[int]]:
    n = max(min(n, len(d)), 1)
    chunks = [list() for _ in range(n)]
    keys = list(d.keys())

    for i, key in enumerate(keys):
        chunks[i % n].append(key)
    return chunks


def count_jsonl_objects(file_path):
    with open(file_path, "r") as file:
        count = sum(1 for _ in file)
    return count


def language_tool_scan(
    instructions: dict, keys: List[int]
) -> Dict[int, InstructionFeature]:
    language_tool = None
    try:
        results = dict()
        language_tool = language_tool_python.LanguageTool(
            "pl-PL",
        )
        for key in keys:
            instruction = instructions[key]
            matches = language_tool.check(instruction.text)
            results[key] = InstructionFeature(
                instruction_id=instruction.id,
                name=LANGUAGE_TOOL_CHECK,
                value=Counter([m.ruleId for m in matches]),
            )

    finally:
        if language_tool:
            language_tool.close()

    return results


def get_samples_per_group(groups: List[Sequence], k: int) -> List[int]:
    num_groups = len(groups)
    ideal_samples_per_group = math.floor(k / num_groups)

    num_per_group = [min(len(g), ideal_samples_per_group) for g in groups]

    total_samples_from_uniform = sum(num_per_group)

    remaining_samples = k - total_samples_from_uniform

    while remaining_samples > 0:
        for i, group in enumerate(groups):
            if num_per_group[i] < len(group) and remaining_samples > 0:
                num_per_group[i] += 1
                remaining_samples -= 1
            if remaining_samples <= 0:
                break

    return num_per_group


def compute_df_row_md5(row):
    str_to_hash = "".join(row)
    result = hashlib.md5(str_to_hash.encode())
    return result.hexdigest()



import os
import random
import shutil
from functools import cached_property
from typing import List, Callable, Optional

import jsonlines
import pandas as pd
import torch
from sentence_transformers import CrossEncoder
from tqdm import tqdm

from elemelek.db import InstructionsDB
from elemelek.features import (
    BasicFeaturesExtractor,
    RerankerRelevanceScoreFeatureExtractor,
    MultiThreadedLanguageToolFeatureExtractor,
)
from elemelek.index import InstructionsSemanticIndex, InstructionsCluster
from elemelek.logging import SelfLogging
from elemelek.model import Instruction, SubsetChoiceMethod
from elemelek.settings import ELEMELEK_ARTIFACTS_PATH, Egg
from elemelek.utils import calculate_file_md5, get_samples_per_group
from transformers import AutoTokenizer


class Elemelek(SelfLogging):
    def __init__(self, config: Egg):
        self.config = config
        self.__dataset_id = calculate_file_md5(config.dataset_jsonl_path)
        os.makedirs(self.dataset_artifacts_path, exist_ok=True)

        self.db = InstructionsDB(self.dataset_artifacts_path)
        self.index = InstructionsSemanticIndex(self.index_dir_path)
        self._clustering = None
        if not os.path.exists(self.config_path):
            shutil.copy(config.path, self.config_path)
            self._build()

    @property
    def config_path(self):
        return os.path.join(ELEMELEK_ARTIFACTS_PATH, self.__dataset_id, "config.yaml")

    @property
    def dataset_artifacts_path(self):
        return os.path.join(ELEMELEK_ARTIFACTS_PATH, self.__dataset_id)

    @property
    def index_dir_path(self):
        return os.path.join(self.dataset_artifacts_path, "index")

    @property
    def clustering(self) -> list[InstructionsCluster]:
        return self.index.cluster(k=self.config.semantic_index.n_clusters)

    @cached_property
    def feature_names(self) -> List[str]:
        return self.db.list_feature_names()

    @staticmethod
    def list_datasets():
        datasets = dict()
        for dataset_id in os.listdir(ELEMELEK_ARTIFACTS_PATH):
            config = Egg.from_yaml(
                os.path.join(ELEMELEK_ARTIFACTS_PATH, dataset_id, "config.yaml")
            )
            datasets[dataset_id] = config
        return datasets

    @staticmethod
    def remove_dataset(dataset_id: str):
        if os.path.exists(os.path.join(ELEMELEK_ARTIFACTS_PATH, dataset_id)):
            shutil.rmtree(os.path.join(ELEMELEK_ARTIFACTS_PATH, dataset_id))

    def _build(self):
        try:
            self.db.load_jsonl(
                self.config.dataset_jsonl_path,
                chunksize=self.config.db.database_insert_batch_size,
            )
            self.index.build(
                self.db,
                model_name=self.config.semantic_index.embeddings_model_name,
                batch_size=self.config.semantic_index.embeddings_computation_batch_size,
                dtype=self.config.semantic_index.dtype,
                metric=self.config.semantic_index.metric,
                connectivity=self.config.semantic_index.connectivity,
                expansion_add=self.config.semantic_index.expansion_add,
                expansion_search=self.config.semantic_index.expansion_search,
            )

            self.index.cluster(k=self.config.semantic_index.n_clusters)
            if self.config.features.basic:
                self.extract_basic_features()
            if self.config.features.reranker:
                self.extract_rerank_relevancy(
                    reranker_model_name=self.config.features.reranker.model_name,
                    reranker_batch_size=self.config.features.reranker.batch_size,
                )
            if self.config.features.language_tool:
                self.extract_lang_mistakes(
                    lang=self.config.features.language_tool.lang,
                    nr_of_threads=self.config.features.language_tool.nr_of_threads,
                )
        except Exception as e:
            Elemelek.remove_dataset(self.__dataset_id)
            raise e

    @classmethod
    def from_dataset_id(cls, dataset_id: str):
        if dataset_id in Elemelek.list_datasets():
            config = Egg.from_yaml(
                os.path.join(ELEMELEK_ARTIFACTS_PATH, dataset_id, "config.yaml")
            )
            return cls(config)

    def extract_basic_features(self):
        extractor = BasicFeaturesExtractor()
        features = extractor.extract_many(self.db)
        self.db.insert_features(features)

    def extract_rerank_relevancy(
        self, reranker_model_name: str, reranker_batch_size: int
    ):
        encoder = CrossEncoder(
            reranker_model_name,
            default_activation_function=torch.nn.Sigmoid(),
            max_length=512,
            device="cuda" if torch.cuda.is_available() else "cpu",
        )
        extractor = RerankerRelevanceScoreFeatureExtractor(
            encoder, batch_size=reranker_batch_size
        )
        features = extractor.extract_many(self.db)
        self.db.insert_features(features)

    def extract_lang_mistakes(self, lang: str, nr_of_threads: int):
        extractor = MultiThreadedLanguageToolFeatureExtractor(
            lang_code=lang, nr_of_threads=nr_of_threads
        )
        features = extractor.extract_many(self.db)
        self.db.insert_features(features)

    def search(self, query: str, k: int = 10) -> List[Instruction]:
        instruction_ids = self.index.search(query, k)
        return [self.db[int(idx)] for idx in instruction_ids]

    def to_pandas(
        self, indices: Optional[List[int]] = None, include_features: bool = True
    ):
        if indices:
            return pd.DataFrame(
                [
                    elem.to_dict(include_features=include_features)
                    for elem in self.db.yield_subset(indices)
                ]
            )
        return pd.DataFrame([elem.to_flat_dict() for elem in self.db])

    def start_sampling(self, shuffle: bool = False):
        return ElemelekSample(self, self.db.ids, shuffle=shuffle)

    def to_jsonl(
        self,
        output_file_path: str,
        indices: Optional[List[int]] = None,
        include_features: bool = False,
        include_instruction_using_chat_template_from: Optional[str] = None,
    ):
        tokenizer = None
        if include_instruction_using_chat_template_from:
            tokenizer = AutoTokenizer.from_pretrained(
                include_instruction_using_chat_template_from
            )
        with jsonlines.open(output_file_path, "w") as jsonl_f:
            if not indices:
                indices = self.db.ids

            for elem in tqdm(
                self.db.yield_subset(indices),
                total=len(indices),
                desc="saving to jsonl",
            ):
                elem_dict = elem.to_dict(include_features=include_features)
                if tokenizer:
                    __instruction_text = tokenizer.apply_chat_template(
                        elem.to_conversation_dict(), tokenize=False
                    )
                    elem_dict |= {"__instruction_text": __instruction_text}
                jsonl_f.write(elem_dict)

    def __getitem__(self, idx: int | slice | list) -> Instruction | List[Instruction]:
        return self.db[idx]


class ElemelekSample(SelfLogging):
    def __init__(self, elemelek: Elemelek, ids: List[int], shuffle: bool = False):
        self.elemelek = elemelek
        self.ids = ids
        if shuffle:
            self.shuffle()

    def shuffle(self) -> "ElemelekSample":
        return ElemelekSample(
            elemelek=self.elemelek, ids=random.sample(self.ids, len(self.ids))
        )

    def unique(self) -> "ElemelekSample":
        return ElemelekSample(elemelek=self.elemelek, ids=self.elemelek.db.unique_ids())

    def filter(
        self, f: Callable[[Instruction], bool], max_k: Optional[int] = None
    ) -> "ElemelekSample":
        filtered_ids = list()
        for instruction in tqdm(
            self.elemelek.db.yield_subset(self.ids),
            desc="Filtering dataset...",
            total=len(self.ids),
        ):
            if f(instruction):
                filtered_ids.append(instruction.id)
            if max_k:
                filtered_ids = filtered_ids[:max_k]
        return ElemelekSample(self.elemelek, filtered_ids)

    def sample_diverse(
        self, k: int, method: SubsetChoiceMethod, **kwargs
    ) -> "ElemelekSample":
        if len(self.ids) <= k:
            self.info(f"Your current subset has less than {k} elements")
            return self
        return_ids = []
        ids_set = set(self.ids)
        clustering = [
            cluster.keep(ids_set)
            for cluster in tqdm(
                self.elemelek.clustering,
                "Filtering clustering to keep sample elements only ...",
            )
        ]
        clustering = [c for c in clustering if len(c) > 0]

        samples_per_cluster = get_samples_per_group(clustering, k)

        if method == SubsetChoiceMethod.RANDOM:
            for i, c in tqdm(enumerate(clustering)):
                return_ids += c.random_sample(samples_per_cluster[i])

        if method == SubsetChoiceMethod.VARIABILITY_FACTOR:
            within_cluster_diversity_factor = kwargs.get(
                "within_cluster_diversity_factor"
            )
            if within_cluster_diversity_factor is None:
                raise ValueError(
                    f"Please provide float within_cluster_diversity_factor "
                    f"parameter when  choosing {method}"
                )
            for i, c in tqdm(
                enumerate(clustering),
                total=len(clustering),
                desc="Solving diverse subset problem heuristically "
                "using genetic algorithms",
            ):
                if samples_per_cluster[i] == 0:
                    continue
                # TODO: multiprocessing approach would help here
                return_ids += c.get_semantically_similar_sample(
                    index=self.elemelek.index,
                    k=samples_per_cluster[i],
                    within_cluster_diversity_factor=within_cluster_diversity_factor,
                )

                assert len(return_ids) == sum(samples_per_cluster[: (i + 1)])

        return ElemelekSample(self.elemelek, return_ids)

    def stratify(self, feature_name: str, k: int) -> "ElemelekSample":
        if len(self.ids) <= k:
            self.info(f"Your current subset has less than {k} elements")
            return self
        values = []
        indices = []
        for instruction in tqdm(
            self.elemelek.db.yield_subset(self.ids),
            desc=f"retrieving values of {feature_name}",
            total=len(self.ids),
        ):
            values.append(instruction.get_feature(feature_name).value)
            indices.append(instruction.id)
        s = pd.Series(values, index=indices)
        groups = [g[1].index.tolist() for g in s.groupby(s)]
        samples_per_group = get_samples_per_group(groups, k)
        sample_ids = []
        for i, sample_size in enumerate(samples_per_group):
            sample_ids += random.sample(groups[i], sample_size)

        return ElemelekSample(self.elemelek, sample_ids)

    def sample_random(self, k: int) -> "ElemelekSample":
        return ElemelekSample(elemelek=self.elemelek, ids=random.sample(self.ids, k=k))

    def __len__(self):
        return len(self.ids)

    def __add__(self, other):
        return ElemelekSample(
            elemelek=self.elemelek, ids=list(set(self.ids) | set(other.ids))
        )

    def to_pandas(self, include_features: bool = True):
        self.elemelek.to_pandas(self.ids, include_features=include_features)

    def to_jsonl(
        self,
        output_file_path: str,
        include_features: bool = False,
        include_instruction_using_chat_template_from: Optional[str] = None,
    ):
        self.elemelek.to_jsonl(
            output_file_path=output_file_path,
            indices=self.ids,
            include_features=include_features,
            include_instruction_using_chat_template_from=include_instruction_using_chat_template_from,
        )

    def __getitem__(
        self, idx: int | slice | list
    ) -> Optional[Instruction | List[Instruction]]:
        if idx in self.ids:
            return self.elemelek[idx]



import json
import os
import sqlite3
import traceback
from typing import List, Dict, Callable, Any, Set

import pandas as pd
from tqdm import tqdm

from elemelek.logging import SelfLogging
from elemelek.model import DatasetError, Instruction, InstructionFeature
from elemelek.settings import MANDATORY_FIELDS
from elemelek.utils import count_jsonl_objects, compute_df_row_md5
from collections.abc import Sequence


class InMemoryFeatures(SelfLogging):
    def __init__(self, features_json_path: str):
        self.features_json_path = features_json_path
        self._feature_names = set()
        self._data = {}
        if os.path.exists(self.features_json_path):
            with open(self.features_json_path, "r") as f:
                _data = json.load(f)
                for k, v in _data.items():
                    self._data[int(k)] = v
                    self._feature_names |= set(v.keys())

        else:
            self._data: Dict[int, Dict[str, InstructionFeature]] = dict()

    def add(self, features: List[InstructionFeature], save: bool = True):
        self._feature_names |= {f.name for f in features}
        for feature in features:
            self._data[feature.instruction_id] = self._data.get(
                feature.instruction_id, dict()
            ) | {feature.name: feature.value}

        if save:
            self.save()

    def filter(self, feature_name, f: Callable[[Any], bool]) -> Set[int]:
        accepted_instuctions = set()
        for instruction_id in self._data:
            if f(self._data[instruction_id][feature_name]):
                accepted_instuctions.add(instruction_id)
        return accepted_instuctions

    def save(self):
        with open(self.features_json_path, "w") as f:
            json.dump(self._data, f)

    def get_features(self, instruction_id: int) -> List[InstructionFeature]:
        return [
            InstructionFeature(instruction_id=instruction_id, name=name, value=value)
            for name, value in self._data.get(instruction_id, dict()).items()
        ]

    def list_feature_names(self) -> List[str]:
        return list(self._feature_names)


class InstructionsDB(SelfLogging, Sequence):
    def __init__(self, data_dir_path: str):
        self.data_dir_path = data_dir_path
        self.conn = sqlite3.connect(self.sqlite_db_path)
        self.dataset_table_name = "dataset"
        self.features = InMemoryFeatures(self.features_json_path)

    @property
    def sqlite_db_path(self):
        return os.path.join(self.data_dir_path, "data.db")

    @property
    def features_json_path(self):
        return os.path.join(self.data_dir_path, "features.json")

    def close(self):
        self.conn.close()

    def load_jsonl(
        self, jsonl_path: str, chunksize: int = 25000, remove_duplicates: bool = True
    ):
        try:
            seen_instructions = set()
            reader = pd.read_json(jsonl_path, lines=True, chunksize=chunksize)
            duplicated = 0
            current_index = 0
            for chunk in tqdm(
                reader,
                total=count_jsonl_objects(jsonl_path) // chunksize,
                desc=f"Inserting chunks of {chunksize} entries to {self.sqlite_db_path}",
            ):
                if MANDATORY_FIELDS.difference(set(chunk.columns)):
                    raise DatasetError(
                        f"Make sure your json entries have all the fields"
                        f": {MANDATORY_FIELDS} "
                    )

                if remove_duplicates:
                    current_chunk_hashes = chunk[list[MANDATORY_FIELDS]].apply(
                        compute_df_row_md5, axis=1
                    )
                    keep_these = current_chunk_hashes.map(
                        lambda x: x not in seen_instructions
                    )
                    seen_instructions |= set(current_chunk_hashes.tolist())

                    duplicated += len(chunk) - keep_these.sum()
                    chunk = chunk[keep_these]

                chunk.set_index(
                    pd.RangeIndex(
                        start=current_index,
                        stop=current_index + len(chunk),
                    ),
                    inplace=True,
                )
                current_index += len(chunk)

                chunk[list(MANDATORY_FIELDS)].to_sql(
                    self.dataset_table_name, self.conn, if_exists="append"
                )

                other_columns = set(chunk.columns) - MANDATORY_FIELDS
                other_features = []
                if len(other_columns) > 0:
                    for column in other_columns:
                        other_features += [
                            InstructionFeature(
                                name=column, instruction_id=id_, value=value
                            )
                            for id_, value in chunk[column].items()
                        ]
                    self.features.add(other_features, save=True)

            self.info(
                f"Data has been saved to database "
                + f"{duplicated} duplicates filtered out"
                if duplicated
                else ""
            )
        except Exception as e:
            self.error(traceback.format_exc())
            self.info(f"removing database {self.sqlite_db_path} due to error")
            os.remove(self.sqlite_db_path)
            raise e

    def __len__(self):
        cur = self.conn.cursor()
        res = cur.execute(f"SELECT count(*) from {self.dataset_table_name}")
        count = res.fetchone()
        return count[0]

    def __iter__(self):
        return self.__yield_all_rows()

    def __getitem__(self, idx: int | slice | list) -> Instruction | List[Instruction]:
        cur = self.conn.cursor()

        if isinstance(idx, slice) or isinstance(idx, list):
            if isinstance(idx, slice):
                indices = list(range(0, len(self)))[idx]
            else:
                indices = idx

            instructions = []
            for i in indices:
                instructions.append(self[i])
            return instructions
        elif isinstance(idx, int):
            res = cur.execute(
                f'SELECT "index", instruction, input, output FROM {self.dataset_table_name} WHERE "index" = {idx}'
            )
            row = res.fetchone()
            if row:
                id_, instruction, input_, output = row
                return Instruction(
                    id=id_,
                    instruction=instruction,
                    input=input_,
                    output=output,
                    features=self.features.get_features(instruction_id=idx),
                )

    def list_feature_names(self) -> List[str]:
        return self.features.list_feature_names()

    def insert_features(self, features: List[InstructionFeature]):
        self.features.add(features)

    def __yield_all_rows(self) -> Instruction:
        cursor = self.conn.cursor()
        cursor.execute(
            f'SELECT "index", instruction, input, output FROM {self.dataset_table_name}'
        )
        row = cursor.fetchone()
        while row:
            if row:
                id_, instruction, input_, output = row
                yield Instruction(
                    id=id_,
                    instruction=instruction,
                    input=input_,
                    output=output,
                    features=self.features.get_features(instruction_id=id_),
                )
            row = cursor.fetchone()
        cursor.close()

    def yield_subset(self, ids: List[int]):
        for id_ in ids:
            yield self[id_]

    @property
    def ids(self) -> List[int]:
        cursor = self.conn.cursor()
        res = cursor.execute(f'SELECT "index" FROM {self.dataset_table_name}')
        rows = res.fetchall()
        return [row[0] for row in rows]

    def unique_ids(self) -> List[int]:
        cursor = self.conn.cursor()
        res = cursor.execute(
            f'SELECT "index" FROM {self.dataset_table_name} group by input, instruction, output;'
        )
        rows = res.fetchall()
        return [row[0] for row in rows]

    def clean(self):
        cur = self.conn.cursor()
        cur.execute("DELETE from dataset;")
        self.conn.commit()



import dataclasses
import os
from distutils.util import strtobool

import yaml
from dotenv import load_dotenv

load_dotenv()
LOGGING_LEVEL = os.environ.get("LOGGING_LEVEL", "DEBUG")
LOGGING_FORMAT = os.environ.get(
    "LOGGING_FORMAT", "[%(asctime)s] [%(levelname)8s] --- %(name)s - %(message)s"
)
LOGGING_PLAIN = strtobool(os.environ.get("LOGGING_PLAIN", "false"))
from pathlib import Path


ELEMELEK_ARTIFACTS_PATH = os.getenv(
    "ELEMELEK_ARTIFACTS_PATH", os.path.join(str(Path.home()), ".elemelek")
)

if not os.path.exists(ELEMELEK_ARTIFACTS_PATH):
    os.makedirs(ELEMELEK_ARTIFACTS_PATH, exist_ok=True)

INSTRUCTION_FIELD = "instruction"
INPUT_FIELD = "input"
OUTPUT_FIELD = "output"
MANDATORY_FIELDS = {INSTRUCTION_FIELD, INPUT_FIELD, OUTPUT_FIELD}


@dataclasses.dataclass
class DBConfig:
    database_insert_batch_size: int
    remove_duplicates: bool


@dataclasses.dataclass
class SemanticIndexConfig:
    embeddings_model_name: str
    embeddings_computation_batch_size: int
    metric: str
    connectivity: int
    dtype: str
    expansion_add: int
    expansion_search: int
    n_clusters: int


@dataclasses.dataclass
class RerankerConfig:
    model_name: str
    batch_size: int
    strategy: str


@dataclasses.dataclass
class LanguageToolConfig:
    nr_of_threads: int
    lang: str


@dataclasses.dataclass
class FeaturesConfig:
    basic: bool
    reranker: RerankerConfig
    language_tool: LanguageToolConfig


@dataclasses.dataclass
class Egg:
    path: str
    dataset_jsonl_path: str
    db: DBConfig
    semantic_index: SemanticIndexConfig
    features: FeaturesConfig

    @classmethod
    def from_yaml(cls, path: str):
        with open(path, "r") as file:
            config_data = yaml.safe_load(file)
        return cls(
            path=path,
            dataset_jsonl_path=config_data["dataset_jsonl_path"],
            db=DBConfig(**config_data["db"]),
            semantic_index=SemanticIndexConfig(**config_data["semantic_index"]),
            features=FeaturesConfig(
                basic=config_data["features"]["basic"],
                reranker=RerankerConfig(**config_data["features"]["reranker"])
                if config_data["features"]["reranker"]
                else None,
                language_tool=LanguageToolConfig(
                    **config_data["features"]["language_tool"]
                )
                if config_data["features"]["language_tool"]
                else None,
            ),
        )



import dataclasses
import json
import os
import random
import shutil

from functools import cached_property
from typing import List, Sequence, Set

import numpy as np
from sentence_transformers import SentenceTransformer

from usearch.index import Index

from elemelek.genetic import OptimalSubsetGeneticAlgorithm
from elemelek.logging import SelfLogging
from elemelek.model import (
    EmbeddingComputationStrategy,
    Instruction,
)
from elemelek.utils import calculate_text_chunk_md5


from tqdm import tqdm


@dataclasses.dataclass
class InstructionsCluster(Sequence, SelfLogging):
    centroid_id: int
    elements_ids: List[int]

    def dict(self):
        return dataclasses.asdict(self)

    def keep(self, ids: Set[int]) -> "InstructionsCluster":
        return InstructionsCluster(
            centroid_id=self.centroid_id,
            elements_ids=[e for e in self.elements_ids if e in ids],
        )

    def random_sample(self, n: int) -> List[int]:
        return random.sample(self.elements_ids, n)

    def get_distance_matrix(self, index: "InstructionsSemanticIndex"):
        distance_matrix = np.zeros(
            shape=(len(self.elements_ids), len(self.elements_ids))
        )
        indices_left = []
        indices_right = []
        left = []
        right = []
        for i, idx_i in enumerate(self.elements_ids):
            for j, idx_j in enumerate(self.elements_ids):
                if i > j:
                    continue
                indices_left.append(i)
                indices_right.append(j)
                left.append(idx_i)
                right.append(idx_j)

        distance_matrix[indices_left, indices_right] = index.index.pairwise_distance(
            left, right
        )
        distance_matrix += distance_matrix.T
        np.fill_diagonal(distance_matrix, 0)
        return distance_matrix

    def get_semantically_similar_sample(
        self,
        index: "InstructionsSemanticIndex",
        k: int,
        within_cluster_diversity_factor: float,
    ) -> List[int]:
        if 0 > k > len(self):
            raise ValueError(
                f"Cannot sample {k} elements from cluster of size {len(self)}"
            )
        if k == 1:
            return [random.choice(self.elements_ids)]
        elif k == len(self):
            return self.elements_ids
        else:
            distance_matrix = self.get_distance_matrix(index)
            target_distance = np.quantile(
                distance_matrix[np.triu(distance_matrix, k=1).nonzero()],
                within_cluster_diversity_factor,
            )

            ga = OptimalSubsetGeneticAlgorithm(
                distance_matrix,
                target_distance=target_distance,
                population_size=100,
                sample_size=k,
                generations=50,
                mutation_rate=0.1,
            )
            solution = ga.optimize()

            optimal_elements_ids = [self.elements_ids[i] for i in solution]

            assert len(optimal_elements_ids) == k
            return optimal_elements_ids

    def __len__(self):
        return len(self.elements_ids)

    def __getitem__(self, index: int) -> int:
        return self.elements_ids[index]


class InstructionsSemanticIndex(SelfLogging):
    def __init__(self, index_dir_path: str):
        self.index_dir_path = index_dir_path
        os.makedirs(index_dir_path, exist_ok=True)
        os.makedirs(self.embeddings_path, exist_ok=True)

        self.__model_name = None
        self.index = None

        if os.path.exists(self.index_path):
            self.info(f"Index at {self.index_path} already exists - loading it")
            self.index = Index.restore(self.index_path)

        if os.path.exists(self.encoder_config_path):
            with open(self.encoder_config_path, "r") as f:
                encoder_data = json.load(f)
                self.__model_name = encoder_data["model_name"]

    @cached_property
    def _encoder(self):
        self.debug(f"Loading {self.__model_name}")
        return SentenceTransformer(self.__model_name)

    @property
    def embeddings_path(self):
        return os.path.join(self.index_dir_path, "embeddings")

    @property
    def index_path(self):
        return os.path.join(self.index_dir_path, "index.bin")

    @property
    def encoder_config_path(self):
        return os.path.join(self.index_dir_path, "encoder.json")

    def build(
        self,
        instructions: Sequence[Instruction],
        model_name: str,
        batch_size: int = 1000,
        force: bool = False,
        **usearch_kwargs,
    ):
        self.__model_name = model_name
        if force or not self.index:
            self._compute_embeddings(
                instructions, EmbeddingComputationStrategy.TRUNCATE, batch_size
            )
            self.index = self._build_index(**usearch_kwargs)
        else:
            self.info(
                f"Index at {self.index_path} already loaded. "
                f"If you want to rebuild pass force = True"
            )

    def cluster(self, k: int = None) -> List[InstructionsCluster]:
        if not self.index:
            raise ValueError("Build  your index first via build()")

        if not k:
            k = len(self.index) // 100
            self.info(f"Nr of clusters not set - k automatically set to {k}")

        if os.path.exists(os.path.join(self.index_dir_path, f"clustering_{k}.json")):
            with open(
                os.path.join(self.index_dir_path, f"clustering_{k}.json"), "r"
            ) as f:
                self.debug(
                    f'Clustering {os.path.join(self.index_dir_path, f"clustering_{k}.json")} already exists'
                )
                clustering = json.load(f)
                return [InstructionsCluster(**c) for c in clustering]
        try:
            clustering = self.index.cluster(
                min_count=k,
                max_count=k,
            )
            centroids, _ = clustering.centroids_popularity
            clusters = []
            for centroid in centroids:
                cluster = InstructionsCluster(
                    centroid_id=int(centroid),
                    elements_ids=[
                        int(i) for i in clustering.members_of(centroid).tolist()
                    ],
                )
                clusters.append(cluster)
        except RuntimeError as e:
            if e.args[0] == "Index too small to cluster!":
                self.warn(
                    "Your index is to small to be clustered - "
                    "returning whole index as one cluster"
                )
                keys = [self.index.keys[k] for k in range(len(self.index))]
                clusters = [InstructionsCluster(centroid_id=keys[0], elements_ids=keys)]
            else:
                raise e
        with open(os.path.join(self.index_dir_path, f"clustering_{k}.json"), "w") as f:
            json.dump([c.dict() for c in clusters], f)
        return clusters

    def _build_index(self, save: bool = True, **usearch_kwargs):
        embeddings_dim = self._encoder.get_sentence_embedding_dimension()

        embeddings_files = os.listdir(self.embeddings_path)

        index = Index(ndim=embeddings_dim, **usearch_kwargs)
        for i, f in tqdm(
            enumerate(embeddings_files),
            total=len(embeddings_files),
            desc="Building HNSW index",
        ):
            data = np.load(os.path.join(self.embeddings_path, f))
            embeddings = data["embeddings"]
            keys = data["keys"]
            index.add(keys=[int(k) for k in keys], vectors=embeddings)

        if save:
            index.save(self.index_path)
            with open(self.encoder_config_path, "w") as f:
                json.dump({"model_name": self.__model_name}, f)
            self.info(f"Index saved at {self.index_path}")
        return index

    def _compute_embeddings(
        self,
        instructions: Sequence[Instruction],
        strategy: EmbeddingComputationStrategy,
        batch_size: int = 1000,
    ):
        if strategy == EmbeddingComputationStrategy.TRUNCATE:
            self._compute_truncated_embeddings(instructions, batch_size)

    def _compute_truncated_embeddings(
        self, instructions: Sequence[Instruction], batch_size: int = 3
    ):
        shutil.rmtree(self.embeddings_path)
        os.makedirs(self.embeddings_path, exist_ok=True)
        embedder = self._encoder
        for start_index in tqdm(
            list(range(0, len(instructions), batch_size)), desc="Computing embeddings"
        ):
            end_index = start_index + batch_size
            batch = instructions[start_index:end_index]
            collection_hash = calculate_text_chunk_md5([b.text for b in batch])
            if os.path.exists(
                os.path.join(self.embeddings_path, f"{collection_hash}.bin")
            ):
                continue
            keys = [b.id for b in batch]
            embeddings = embedder.encode([b.text for b in batch])

            with open(
                os.path.join(self.embeddings_path, f"{collection_hash}.bin"), "wb"
            ) as fp:
                np.savez_compressed(fp, keys=keys, embeddings=embeddings)

    def search(self, query: str, k: int = 2) -> List[int]:
        encoded_query = self._encoder.encode(query)
        matches = self.index.search(encoded_query, count=k)
        return [m.key for m in matches]



from abc import ABC, abstractmethod
from typing import List, Sequence

import numpy as np
from sentence_transformers import CrossEncoder
from torch import nn
from tqdm import tqdm

from elemelek.logging import SelfLogging
from elemelek.model import (
    Instruction,
    InstructionFeature,
    CustomExtractorDefinition,
)


from elemelek.utils import divide_dict, language_tool_scan
import concurrent.futures


RERANKER_RELEVANCE_SCORE = "reranker-relevance-score"
LANGUAGE_TOOL_CHECK = "language_tool_check"
MEDIAN_WORD_LENGTH = "median_word_length"
QUANTILE_WORD_LENGTH_09 = "quantile_0.9_word_length"
QUANTILE_WORD_LENGTH_01 = "quantile_0.1_word_length"
TOTAL_LENGTH = "total_length"
IS_QUESTION = "is_question"
HAS_INPUT = "has_input"
NUMERIC_CHARS_RATIO = "numeric_chars_ratio"
NON_ALPHA_NUMERIC_CHARS_RATIO = "non_alpha_numeric_chars_ratio"


class InstructionFeatureExtractor(ABC, SelfLogging):
    @abstractmethod
    def extract(self, instruction: Instruction) -> List[InstructionFeature]:
        pass

    def extract_many(
        self, instructions: Sequence[Instruction]
    ) -> List[InstructionFeature]:
        features = []
        for instruction in tqdm(
            instructions, desc=f"Extracting features via {self.__class__.__name__}..."
        ):
            features += self.extract(instruction)

        return features


class BasicFeaturesExtractor(InstructionFeatureExtractor):
    def extract(self, instruction: Instruction) -> List[InstructionFeature]:
        words = instruction.text.split()
        word_lengths = [len(word) for word in words]
        non_alpha_num_chars = sum(1 for char in instruction.text if not char.isalnum())
        numeric_chars = sum(1 for char in instruction.text if char.isdigit())

        return [
            InstructionFeature(
                name=MEDIAN_WORD_LENGTH,
                value=np.median(word_lengths),
                instruction_id=instruction.id,
            ),
            InstructionFeature(
                name=QUANTILE_WORD_LENGTH_09,
                value=np.quantile(word_lengths, 0.9),
                instruction_id=instruction.id,
            ),
            InstructionFeature(
                name=QUANTILE_WORD_LENGTH_01,
                value=np.quantile(word_lengths, 0.1),
                instruction_id=instruction.id,
            ),
            InstructionFeature(
                name=TOTAL_LENGTH,
                value=len(instruction.text),
                instruction_id=instruction.id,
            ),
            InstructionFeature(
                name=IS_QUESTION,
                value=instruction.is_question(),
                instruction_id=instruction.id,
            ),
            InstructionFeature(
                name=HAS_INPUT,
                value=instruction.input is not None,
                instruction_id=instruction.id,
            ),
            InstructionFeature(
                name=NUMERIC_CHARS_RATIO,
                value=numeric_chars / len(instruction.text),
                instruction_id=instruction.id,
            ),
            InstructionFeature(
                name=NON_ALPHA_NUMERIC_CHARS_RATIO,
                value=non_alpha_num_chars / len(instruction.text),
                instruction_id=instruction.id,
            ),
        ]


# https://stackoverflow.com/a/73464315
class MultiThreadedLanguageToolFeatureExtractor(InstructionFeatureExtractor):
    def __init__(self, lang_code: str = "pl-PL", nr_of_threads: int = 4):
        self.lang_code = lang_code
        self.nr_of_threads = nr_of_threads

    def extract(self, instruction: Instruction) -> InstructionFeature:
        raise NotImplemented(
            "single instruction extract is not supported in multithreaded setup"
        )

    def extract_many(
        self, instructions: Sequence[Instruction]
    ) -> List[InstructionFeature]:
        final_results = {}
        instructions_dict = {
            instruction.id: instruction for instruction in instructions
        }  # This bit sucks - it loads everything into memory 😬💀
        chunks = divide_dict(instructions_dict, self.nr_of_threads)

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.nr_of_threads
        ) as executor:
            futures = [
                executor.submit(language_tool_scan, instructions_dict, chunk)
                for chunk in chunks
            ]
        for future in concurrent.futures.as_completed(futures):
            final_results.update(future.result())

        return [final_results[instruction.id] for instruction in instructions]


class CustomFeatureExtractor(InstructionFeatureExtractor):
    def __init__(self, custom_extractors: List[CustomExtractorDefinition]):
        self.customer_extractors = custom_extractors

    def extract(self, instruction: Instruction) -> List[InstructionFeature]:
        return [extractor.f(instruction) for extractor in self.customer_extractors]


class RerankerRelevanceScoreFeatureExtractor(InstructionFeatureExtractor):
    def __init__(self, reranker: CrossEncoder, batch_size: int = 16):
        self.reranker = reranker
        self.batch_size = batch_size

    def extract(self, instruction: Instruction) -> InstructionFeature:
        results = self.reranker.predict(
            [[instruction.qa.question, instruction.qa.answer]],
            activation_fct=nn.Sigmoid(),
        )
        return InstructionFeature(
            instruction_id=instruction.id,
            name=RERANKER_RELEVANCE_SCORE,
            value=float(results[0]),
        )

    def extract_many(
        self, instructions: Sequence[Instruction]
    ) -> List[InstructionFeature]:
        results = []
        for start_index in tqdm(
            list(range(0, len(instructions), self.batch_size)),
            desc=f"Computing rerank relevance score via {self.reranker.model.config._name_or_path}",
        ):
            end_index = start_index + self.batch_size
            batch = instructions[start_index:end_index]
            batch_results = self.reranker.predict(
                [
                    [instruction.qa.question, instruction.qa.answer]
                    for instruction in batch
                ],
                activation_fct=nn.Sigmoid(),
            )

            results += [
                InstructionFeature(
                    name=RERANKER_RELEVANCE_SCORE,
                    value=float(batch_results[i]),
                    instruction_id=b.id,
                )
                for i, b in enumerate(batch)
            ]

        return results
